/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: fourg.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: fourg.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $ */
#include <math.h>
#include <stdio.h>
void fourg (float *data, int n, int isign, float *work)
/*-----------------------------------------------------------------------
 *     Cooley-Tukey fast fourier transform.
 *     One-dimensional transform of complex data, arbitrary number of
 *     points.  N points can be transformed in time proportional to
 *     N*log(N) (for N non-prime), whereas other methods take N**2 time.
 *     furthermore, because fewer arithmetic operations are performed,
 *     less error is built up.
 *     The transform done is--
 *           Real*4 data(n),transform(n),work(n)
 *            *omplex data,transform,work
 *       transform(k) = sum(data(j)*exp(isign*2*pi*i*(j-1)*(k-1)/n)),
 *     summed from j = 1 to n for all k from 1 to N.  The transform
 *     values are returned to data, replacing the input.  N may be any
 *     positive number, but it should be non-prime for speed.
 *     ISIGN = +1 or -1.
 *     A -1 transform followed by a +1 one (or vice versa)
 *     returns n times the original data.  work is a one-dimensional
 *     complex array of length n used for working storage.
 *     running time is proportional to n * (sum of the prime factors of
 *     n).  for example, n = 1960, time is to * 1960 * (2+2+2+5+7+7).
 *     naive methods directly implementing the summation run in time
 *     proportional to n**2.  an upper bound for the rms relative error
 *     is 3 * 2**(-b) * sum(f**1.5), where b is the number of bits in
 *     the floating point fraction and the sum is over the prime
 *     factors of n.
 *        Note: the input data array uses the, "center at the edges"
 *     convention for data storage, where the first element is zero.
 *     Some rearraingment of the data before and after the FFT may be 
 *     required.
 *
 *  Inputs:
 *     data      Array of length N to be FFT'd
 *     n         Length of data array - need not be a power of two.
 *     isign     Direction of transform
 *     work      Array of length N for temporary storage.
 *
 *  Outputs:
 *     data      Transformed array
 *
 *     Written by Norman Brenner, Mit, August 1968.
 *     Converted to c by W. D. Cotton, October 1999.
 *---------------------------------------------------------------------*/
{
  float twopi, theta, sinth, rootr, rooti, wstpr, wstpi, wminr, wmini;
  float tempr, tempi, wr, wi, sumr, sumi, temp;
  long  ifact[32], jf, npart, id, idiv, iquot;
  long  nfact, ip0, ip3, iwork, i3rev, i3, ip2, ip1;
  long  ifcur, i1, j0, j1, iwmax, i2max, i2;
/*--------------------------------------------------------------------*/
  twopi = 6.2831853072 * (float)isign;
  /* factor n into its prime factors, nfact in number.   */
  /* for example, for n = 1960, nfact = 6 and            */
  /* ifact(if) = 2, 2, 2, 5, 7 and 7.                    */
  for (jf=0; jf<32;jf++) ifact[jf] = 0;
  jf = 0;
  npart = n;
  for (id=1; id<=n; id+=2) {
    idiv = id;
    if (id<=1) idiv = 2;
    while (1) {
      iquot = npart / idiv;
      if (npart != idiv * iquot) break;
      jf = jf + 1;
      ifact[jf-1] = idiv;
      npart = iquot;
    } 
    if (iquot <= idiv) break;
  }
  if (npart > 1) ifact[(++jf)-1] = npart;
  nfact = jf;

  /*  shuffle the data array by reversing the digits of the   */
  /*  index. Replace data(i) by data(irev) for all i from 1   */
  /*  to n.  irev-1 is the integer whose digit representation */
  /*  in the multi-radix notation of factors ifact(if) is the */
  /*  reverse of the representatn of i-1. For example, if all */
  /*  ifact(if) = 2, then for i-1 = 11001, irev-1 = 10011. A  */
  /*  work array of length n is needed.                       */
  ip0 = 2;
  ip3 = ip0 * n;
  iwork = 1;
  i3rev = 1;
  for (i3=1; i3<=ip3; i3+=ip0) { /* 110 loop */
    /* debug 
       fprintf (stderr,"%d i3rev =%d %o %o\n",i3, i3rev,i3, i3rev);*/
    work[iwork-1] = data[i3rev-1];
    work[iwork] = data[i3rev];
    ip2 = ip3;
    for (jf=0; jf<nfact; jf++) { /* 100 loop */
      ip1 = ip2 / ifact[jf];
      i3rev = i3rev + ip1;
      if (i3rev <= ip2) break;
      i3rev = i3rev - ip2;
      ip2 = ip1;
    } /* end 100 loop */
    iwork = iwork + ip0;
  } /* end 110 loop */
  iwork = 1;
  for (i3=1; i3<=ip3; i3+=ip0) { /* 120 loop */
    data[i3-1] = work[iwork-1];
    data[i3] = work[iwork];
    iwork = iwork + ip0;
  } /* end 120 loop */

  /* phase-shifted fourier transform of length ifact[if-1].          */
  /* iprod=ip1/ip0 irem=n/(ifact[if-1]*iprod)                        */
  /* real data[iprod-1,ifact[if-1]-1,irem-1), work[ifact[if-1]]      */
  /* complex data,work                                               */
  /* data[i1-1,j2-1,i3-1] =  sum(data[i1-1,i2-1,i3-1]) * w**(i2-1)), */
  /* summed over i2 = 1 to ifact[if-1] for all i1                    */
  /* from 1 to iprod, j2 from 1 to ifact[if-1] and i3                */
  /* from 1 to irem.                                                 */
  /* w = exp(isign*2*pi*i*(i1-1+iprod*(j2-1))/(iprod*ifact[if-1])).  */
  jf = 0;
  ip1 = ip0;
  while (ip1<ip3) { /* 130 loop */
    jf = jf + 1;
    ifcur = ifact[jf-1];
    ip2 = ip1 * ifcur;
    theta = twopi / ((float)ifcur);
    sinth = sin (theta*0.5);
    rootr = -2. * sinth * sinth;
    rooti = sin (theta);
    theta = twopi / ((float) (ip2/ip0));
    sinth = sin (theta * 0.5);
    wstpr = -2. * sinth * sinth;
    wstpi = sin (theta);
    wminr = 1.;
    wmini = 0.;
    for (i1=1; i1<=ip1; i1+=ip0) { /* 230 loop */
      if (ifcur <= 2 ) {
	for (i3=i1; i3<=ip3; i3+=ip2) { /* loop 160 */
	  j0 = i3;
	  j1 = i3 + ip1;
	  tempr = wminr * data[j1-1] - wmini * data[j1];
	  tempi = wminr * data[j1] + wmini * data[j1-1];
	  data[j1-1] = data[j0-1]  - tempr;
	  data[j1]   = data[j0]    - tempi;
	  data[j0-1] = data[j0-1]  + tempr;
	  data[j0]   = data[j0]    + tempi;
	} /* end loop 160 */
      } else {
	iwmax = ip0 * ifcur;
	for (i3=i1; i3<=ip3; i3+=ip2) { /* loop 210 outer*/
	  i2max = i3 + ip2 - ip1;
	  wr = wminr;
	  wi = wmini;
	  for (iwork=1; iwork<=iwmax; iwork+=ip0) { /* loop 200 */
	    i2 = i2max;
	    sumr = data[i2-1];
	    sumi = data[i2];
	    while (1) { /* start of 180 loop */
	      i2 = i2 - ip1;
	      tempr = sumr;
	      sumr = wr * sumr - wi * sumi + data[i2-1];
	      sumi = wr * sumi + wi * tempr + data[i2];
	      if (i2 <= i3) break;
	    } /* end loop 180 */ 
	    work[iwork-1] = sumr;
	    work[iwork] = sumi;
	    temp = wr;
	    wr = wr * rootr - wi * rooti + wr;
	    wi = temp * rooti + wi * rootr + wi;
	  } /* end loop 200 */
	  iwork = 1;
	  for (i2=i3; i2<=i2max; i2+=ip1) { /* loop 210 inner */
	    data[i2-1] = work[iwork-1];
	    data[i2]   = work[iwork];
	    iwork = iwork + ip0;
	  } /* end loop 210 inner */
	} /* end loop 210 outer*/
      } /* end of else block */
      tempr = wminr;
      wminr = wminr * wstpr - wmini * wstpi + wminr;
      wmini = tempr * wstpi + wmini * wstpr + wmini;
    } /* end 230 loop */
    ip1 = ip2;
  } /* end 130 loop */
} /* end fourg */
/* */
